#include "cdn-instruction-matrix.h"
#include <string.h>

#define CDN_INSTRUCTION_MATRIX_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE((object), CDN_TYPE_INSTRUCTION_MATRIX, CdnInstructionMatrixPrivate))

struct _CdnInstructionMatrixPrivate
{
	CdnStackManipulation smanip;
};

G_DEFINE_TYPE (CdnInstructionMatrix, cdn_instruction_matrix, CDN_TYPE_INSTRUCTION)

static CdnMiniObject *
cdn_instruction_matrix_copy (CdnMiniObject *object)
{
	CdnMiniObject *ret;
	CdnInstructionMatrix *self;
	CdnInstructionMatrix const *src;

	src = CDN_INSTRUCTION_MATRIX_CONST (object);

	ret = CDN_MINI_OBJECT_CLASS (cdn_instruction_matrix_parent_class)->copy (object);

	self = CDN_INSTRUCTION_MATRIX (ret);

	cdn_stack_manipulation_copy (&self->priv->smanip, &src->priv->smanip);

	return ret;
}

static gchar *
cdn_instruction_matrix_to_string (CdnInstruction *instruction)
{
	CdnInstructionMatrix *self;

	self = CDN_INSTRUCTION_MATRIX (instruction);

	return g_strdup_printf ("MAT (%d, %d)",
	                        self->priv->smanip.push.rows,
	                        self->priv->smanip.push.columns);
}

static void
cdn_instruction_matrix_execute (CdnInstruction *instruction,
                                CdnStack       *stack)
{
	// Matrix is a dummy operation
}

static CdnStackManipulation const *
cdn_instruction_matrix_get_stack_manipulation (CdnInstruction  *instruction,
                                               GError         **error)
{
	CdnInstructionMatrix *self;

	self = CDN_INSTRUCTION_MATRIX (instruction);
	return &self->priv->smanip;
}

static gboolean
check_pop_dims (CdnInstructionMatrix *m1,
                CdnInstructionMatrix *m2)
{
	gint i;

	if (m1->priv->smanip.pop.num != m2->priv->smanip.pop.num)
	{
		return FALSE;
	}

	for (i = 0; i < m1->priv->smanip.pop.num; ++i)
	{
		if (m1->priv->smanip.pop.args[i].rows != m2->priv->smanip.pop.args[i].rows ||
		    m1->priv->smanip.pop.args[i].columns != m2->priv->smanip.pop.args[i].columns)
		{
			return FALSE;
		}
	}

	return TRUE;
}

static gboolean
cdn_instruction_matrix_equal (CdnInstruction *i1,
                              CdnInstruction *i2,
                              gboolean        asstring)
{
	CdnInstructionMatrix *n1 = CDN_INSTRUCTION_MATRIX (i1);
	CdnInstructionMatrix *n2 = CDN_INSTRUCTION_MATRIX (i2);

	return n1->priv->smanip.push.rows == n2->priv->smanip.push.rows &&
	       n1->priv->smanip.push.columns == n2->priv->smanip.push.columns &&
	       check_pop_dims (n1, n2);
}

static void
cdn_instruction_matrix_finalize (CdnMiniObject *object)
{
	CdnInstructionMatrix *self;

	self = CDN_INSTRUCTION_MATRIX (object);

	cdn_stack_manipulation_destroy (&self->priv->smanip);

	CDN_MINI_OBJECT_CLASS (cdn_instruction_matrix_parent_class)->finalize (object);
}

static void
cdn_instruction_matrix_recalculate_sparsity (CdnInstruction *instruction)
{
	CdnInstructionMatrix *self;
	CdnStackArgs *args;
	gint i;
	gint sp = 0;

	self = CDN_INSTRUCTION_MATRIX (instruction);

	args = &self->priv->smanip.pop;

	for (i = 0; i < args->num; ++i)
	{
		sp += args->args[i].num_sparse;
	}

	if (sp > 0)
	{
		gint off = 0;
		guint *sparsity = g_new0 (guint, sp);
		gint spi = 0;

		for (i = args->num - 1; i >= 0; --i)
		{
			gint j;

			for (j = 0; j < args->args[i].num_sparse; ++j)
			{
				sparsity[spi++] = args->args[i].sparsity[j] + off;
			}

			off += cdn_dimension_size (&args->args[i].dimension);
		}

		cdn_stack_arg_set_sparsity (&self->priv->smanip.push, sparsity, sp);
		g_free (sparsity);
	}

}

static void
cdn_instruction_matrix_class_init (CdnInstructionMatrixClass *klass)
{
	CdnMiniObjectClass *object_class = CDN_MINI_OBJECT_CLASS (klass);
	CdnInstructionClass *inst_class = CDN_INSTRUCTION_CLASS (klass);

	object_class->finalize = cdn_instruction_matrix_finalize;

	object_class->copy = cdn_instruction_matrix_copy;

	inst_class->to_string = cdn_instruction_matrix_to_string;
	inst_class->execute = cdn_instruction_matrix_execute;
	inst_class->get_stack_manipulation = cdn_instruction_matrix_get_stack_manipulation;
	inst_class->equal = cdn_instruction_matrix_equal;
	inst_class->recalculate_sparsity = cdn_instruction_matrix_recalculate_sparsity;

	g_type_class_add_private (object_class, sizeof(CdnInstructionMatrixPrivate));
}

static void
cdn_instruction_matrix_init (CdnInstructionMatrix *self)
{
	self->priv = CDN_INSTRUCTION_MATRIX_GET_PRIVATE (self);
}

CdnInstruction *
cdn_instruction_matrix_new (CdnStackArgs const *args,
                            CdnDimension const *dim)
{
	CdnMiniObject *ret;
	CdnInstructionMatrix *self;

	ret = cdn_mini_object_new (CDN_TYPE_INSTRUCTION_MATRIX);
	self = CDN_INSTRUCTION_MATRIX (ret);

	cdn_stack_args_copy (&self->priv->smanip.pop, args);
	self->priv->smanip.push.dimension = *dim;

	cdn_instruction_matrix_recalculate_sparsity (CDN_INSTRUCTION (self));

	return CDN_INSTRUCTION (ret);
}
